import itertools
import numpy as np


def get_new_vertices(
    addends: np.ndarray,
    parent: np.ndarray
) -> np.ndarray:
    """
    Parameters
    ----------
    `addends` - `np.ndarray` of shape `(a, d)`

    `parent` - `np.ndarray` of shape `(n, d)`
        In the bonus version, the shape is
        `(b_1,...,b_k, n, d)`

    Returns
    -------
    `new_vertices` - `np.ndarray` of shape (a * n, d)
        This array has as rows the entries `u + v` where
        `u` is a row of `addends` and
        `v` is a row of `parent`.

        In the bonus version,
        `new_vertices` has shape `(b_1,...,b_k,a * n, d)`
        and for `0 <= i_j < b_j` for `1 <= j <= k`,
        the 2-array `new_vertices[i_1,...,i_k]` is equal to
        `get_new_vertices(addends, parent[i_1,...,i_k])`.
    """
    raise NotImplementedError("TODO")


def get_dedupe_mask(
    new_vertices: np.ndarray,
    parent: np.ndarray
) -> np.ndarray:
    """
    Parameters
    ----------
    `new_vertices` - `np.ndarray` of shape `(m, d)`
        In the bonus version, the shape is
        `(b_1,...,b_k, m, d)`

    `parent` - `np.ndarray` of shape `(n, d)`
        In the bonus version, the shape is
        `(b_1,...,b_k, n, d)`

    Returns
    -------
    `dedupe_mask` - Boolean `np.ndarray` of shape `(m,)`
        For `0 <= i < m`, the Boolean value `dedupe_mask[i]`
        says if `new_vertices[i]` is not a row of
        `parent` of `new_vertices[:i]`.

        In the bonus version,
        `dedupe_mask has shape `(b_1,...,b_k, m)`
        and for `0 <= i_j < b_j` for `1 <= j <= k`,
        the 1-array `dedupe_mask[i_1,...,i_k]` is equal to
        `get_dedupe_mask(new_vertices[i_1,...,i_k], parent[i_1,...,i_k])`.
    """
    raise NotImplementedError("TODO")
    


def test_get_new_vertices(bonus=False):
    """
    Parameters
    ----------
    `bonus` - `bool`, default: `False`
        Whether to test batched inputs for the bonus version.
    """
    for addends, parent, new_vertices_wanted in zip(
        ADDENDS,
        PARENT,
        NEW_VERTICES
    ):
        if bonus:
            new_vertices_gotten = get_new_vertices(addends, parent)
            if new_vertices_gotten.shape != new_vertices_wanted.shape:
                raise ValueError(
                    f"In case `addends={addends}` and `parent={parent}`, `new_vertices` has shape {new_vertices_gotten.shape} but it should have shape {new_vertices_wanted.shape}"
                )
            
            new_vertices_gotten = _get_subarrays(new_vertices_gotten, 2)

        else:
            new_vertices_gotten = [
                get_new_vertices(addends, _parent)
                for _parent in _get_subarrays(parent, 2)
            ]

        new_vertices_wanted = _get_subarrays(new_vertices_wanted, 2)

        for _new_vertices_gotten, _new_vertices_wanted, _parent in zip(
            new_vertices_gotten,
            new_vertices_wanted,
            _get_subarrays(parent, 2)
        ):
            if _new_vertices_gotten.shape != _new_vertices_wanted.shape:
                raise ValueError(
                    f"In case `addends={addends}` and `parent={_parent}`, `new_vertices` has shape {_new_vertices_gotten.shape} but it should have shape {_new_vertices_wanted.shape}"
                )
            
            gotten_sorted, wanted_sorted = (
                a[np.lexsort(a.T)]
                for a in (_new_vertices_gotten, _new_vertices_wanted)
            )

            if not np.array_equal(gotten_sorted, wanted_sorted):
                raise ValueError(
                    f"In case `addends={addends}` and `parent={_parent}`, the rows of `new_vertices` in lexicographical order are {gotten_sorted} but they should be {wanted_sorted}"
                )
            
    print(f"All {'bonus' if bonus else 'normal'} tests for `get_new_vertices` passed. Congratulations!")


def test_get_dedupe_mask(bonus=False):
    """
    Parameters
    ----------
    `bonus` - `bool`, default: `False`
        Whether to test batched inputs for the bonus version.
    """
    for new_vertices, parent, dedupe_mask_wanted in zip(
        NEW_VERTICES,
        PARENT,
        DEDUPE_MASK
    ):
        if bonus:
            dedupe_mask_gotten = get_dedupe_mask(new_vertices, parent)
            if dedupe_mask_gotten.shape != dedupe_mask_wanted.shape:
                raise ValueError(
                    f"In case `new_vertices={new_vertices}` and `parent={parent}`, `dedupe_mask` has shape {dedupe_mask_gotten.shape} but it should have shape {dedupe_mask_wanted.shape}"
                )
            
            dedupe_mask_gotten = _get_subarrays(dedupe_mask_gotten, 1)

        else:
            dedupe_mask_gotten = [
                get_dedupe_mask(_new_vertices, _parent)
                for _new_vertices, _parent in zip(*map(
                    lambda a: _get_subarrays(a, 2),
                    [new_vertices, parent]
                ))
            ]

        dedupe_mask_wanted = _get_subarrays(dedupe_mask_wanted, 1)

        for (
            _dedupe_mask_gotten, _dedupe_mask_wanted, _new_vertices, _parent
        ) in zip(
            dedupe_mask_gotten,
            dedupe_mask_wanted,
            _get_subarrays(new_vertices, 2),
            _get_subarrays(parent, 2)
        ):
            if _dedupe_mask_gotten.shape != _dedupe_mask_wanted.shape:
                raise ValueError(
                    f"In case `new_vertices={_new_vertices}` and `parent={_parent}`, `dedupe_mask` has shape {_dedupe_mask_gotten.shape} but it should have shape {_dedupe_mask_wanted.shape}"
                )
            
            if not np.array_equal(_dedupe_mask_gotten, _dedupe_mask_wanted):
                raise ValueError(
                    f"In case `new_vertices={_new_vertices}` and `parent={_parent}`, `dedupe_mask` is {_dedupe_mask_gotten} but it should be {_dedupe_mask_wanted}"
                )
            
    print(f"All {'bonus' if bonus else 'normal'} tests for `get_dedupe_mask` passed. Congratulations!")


ADDENDS = [
    np.array([[1], [-1]]),
    np.array([[1, 0], [-1, 0], [0, 1], [0, -1]]),
    np.array([[1, 0, 0, 0], [-1, 0, 0, 0], [0, 1, 0, 0], [0, -1, 0, 0], [0, 0, 1, 0], [0, 0, -1, 0], [0, 0, 0, 1], [0, 0, 0, -1]])
]
PARENT = [
    np.array([[[0], [1], [2]], [[0], [2], [4]], [[-5], [-7], [7]]]),
    np.array([
        [
            [[0, 0], [1, 0], [2, 0]],
            [[-5, 0], [2, 1], [0, 3]]
        ],
        [
            [[0, 0], [1, 0], [0, 1]],
            [[1, 1], [2, 1], [1, 2]]
        ]
    ]),
    np.array([[0, 0, 0, 0], [1, 0, 0, 0], [0, 1, 0, 0], [1, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1], [0, 0, 1, 1]])
]
NEW_VERTICES = [
    np.array([[[ 1],
                [ 2],
                [ 3],
                [-1],
                [ 0],
                [ 1]],

            [[ 1],
                [ 3],
                [ 5],
                [-1],
                [ 1],
                [ 3]],

            [[-4],
                [-6],
                [ 8],
                [-6],
                [-8],
                [ 6]]]),
    np.array([[[[ 1,  0],
                    [ 2,  0],
                    [ 3,  0],
                    [-1,  0],
                    [ 0,  0],
                    [ 1,  0],
                    [ 0,  1],
                    [ 1,  1],
                    [ 2,  1],
                    [ 0, -1],
                    [ 1, -1],
                    [ 2, -1]],

                    [[-4,  0],
                    [ 3,  1],
                    [ 1,  3],
                    [-6,  0],
                    [ 1,  1],
                    [-1,  3],
                    [-5,  1],
                    [ 2,  2],
                    [ 0,  4],
                    [-5, -1],
                    [ 2,  0],
                    [ 0,  2]]],


                [[[ 1,  0],
                    [ 2,  0],
                    [ 1,  1],
                    [-1,  0],
                    [ 0,  0],
                    [-1,  1],
                    [ 0,  1],
                    [ 1,  1],
                    [ 0,  2],
                    [ 0, -1],
                    [ 1, -1],
                    [ 0,  0]],

                    [[ 2,  1],
                    [ 3,  1],
                    [ 2,  2],
                    [ 0,  1],
                    [ 1,  1],
                    [ 0,  2],
                    [ 1,  2],
                    [ 2,  2],
                    [ 1,  3],
                    [ 1,  0],
                    [ 2,  0],
                    [ 1,  1]]]]),
    np.array([[ 1,  0,  0,  0],
                    [ 2,  0,  0,  0],
                    [ 1,  1,  0,  0],
                    [ 2,  1,  0,  0],
                    [ 1,  0,  1,  0],
                    [ 1,  0,  0,  1],
                    [ 1,  0,  1,  1],
                    [-1,  0,  0,  0],
                    [ 0,  0,  0,  0],
                    [-1,  1,  0,  0],
                    [ 0,  1,  0,  0],
                    [-1,  0,  1,  0],
                    [-1,  0,  0,  1],
                    [-1,  0,  1,  1],
                    [ 0,  1,  0,  0],
                    [ 1,  1,  0,  0],
                    [ 0,  2,  0,  0],
                    [ 1,  2,  0,  0],
                    [ 0,  1,  1,  0],
                    [ 0,  1,  0,  1],
                    [ 0,  1,  1,  1],
                    [ 0, -1,  0,  0],
                    [ 1, -1,  0,  0],
                    [ 0,  0,  0,  0],
                    [ 1,  0,  0,  0],
                    [ 0, -1,  1,  0],
                    [ 0, -1,  0,  1],
                    [ 0, -1,  1,  1],
                    [ 0,  0,  1,  0],
                    [ 1,  0,  1,  0],
                    [ 0,  1,  1,  0],
                    [ 1,  1,  1,  0],
                    [ 0,  0,  2,  0],
                    [ 0,  0,  1,  1],
                    [ 0,  0,  2,  1],
                    [ 0,  0, -1,  0],
                    [ 1,  0, -1,  0],
                    [ 0,  1, -1,  0],
                    [ 1,  1, -1,  0],
                    [ 0,  0,  0,  0],
                    [ 0,  0, -1,  1],
                    [ 0,  0,  0,  1],
                    [ 0,  0,  0,  1],
                    [ 1,  0,  0,  1],
                    [ 0,  1,  0,  1],
                    [ 1,  1,  0,  1],
                    [ 0,  0,  1,  1],
                    [ 0,  0,  0,  2],
                    [ 0,  0,  1,  2],
                    [ 0,  0,  0, -1],
                    [ 1,  0,  0, -1],
                    [ 0,  1,  0, -1],
                    [ 1,  1,  0, -1],
                    [ 0,  0,  1, -1],
                    [ 0,  0,  0,  0],
                    [ 0,  0,  1,  0]])
]
DEDUPE_MASK = [
    np.array([[False, False,  True,  True, False, False],
       [ True,  True,  True,  True, False, False],
       [ True,  True,  True, False,  True,  True]]),
    np.array([[[False, False,  True,  True, False, False,  True,  True,  True,
          True,  True,  True],
        [ True,  True,  True,  True,  True,  True,  True,  True,  True,
          True,  True,  True]],

       [[False,  True,  True,  True, False,  True, False, False,  True,
          True,  True, False],
        [False,  True,  True,  True, False,  True, False, False,  True,
          True,  True, False]]]),
    np.array([False,  True, False,  True,  True,  True,  True,  True, False,
        True, False,  True,  True,  True, False, False,  True,  True,
        True,  True,  True,  True,  True, False, False,  True,  True,
        True, False, False, False,  True,  True, False,  True,  True,
        True,  True,  True, False,  True, False, False, False, False,
        True, False,  True,  True,  True,  True,  True,  True,  True,
       False, False])
]


def _get_subarrays(a: np.ndarray, dims: int) -> list[np.ndarray]:
    """
    Parameters
    ----------
    `a` - `np.ndarray` of shape `(n_1, ..., n_k)`
    `dims` - `int`

    Returns
    -------
    `subarrays` - `list[np.ndarray]`
        a list of the subarrays `a[i_1,...,i_{k-dim}]`
        where `0 <= i_j < n_j` for `1 <= j <= k - dims`.
    """
    subarrays = [a]
    while len(subarrays[0].shape) > dims:
        subarrays = list(itertools.chain(*subarrays))

    return subarrays


if __name__ == "__main__":
    test_get_new_vertices()
    test_get_dedupe_mask()